#ifndef FTPP_BITMAP_H
#define FTPP_BITMAP_H

#include <cassert>
#include <utility>
#include <freetype2/ft2build.h>
#include FT_FREETYPE_H
#include FT_BITMAP_H
#include <ftpp/exception.hpp>

namespace freetype {

class Bitmap {
   public:
      Bitmap() noexcept : library(nullptr), bitmap{0} {
      }
      Bitmap(FT_Library library, FT_Bitmap* bitmap_ptr) :
	    library(library), bitmap{0} {
	 FT_Error error = FT_Bitmap_Copy(library, bitmap_ptr, &bitmap);
	 if (error) throw Exception(error, "FT_Bitmap_Copy");
      }
      Bitmap(const Bitmap& other) : library(other.library), bitmap{0} {
	 FT_Error error = FT_Bitmap_Copy(library, &other.bitmap, &bitmap);
	 if (error) throw Exception(error, "FT_Bitmap_Copy");
      }
      Bitmap(Bitmap&& other) noexcept {
	 swap(other);
      }
      ~Bitmap() {
	 if (library) FT_Bitmap_Done(library, &bitmap);
      }
      Bitmap& operator=(const Bitmap& other) {
	 if (library) {
	    FT_Bitmap_Done(library, &bitmap); library = nullptr;
	 }
	 FT_Error error = FT_Bitmap_Copy(other.library, &other.bitmap, &bitmap);
	 if (error) throw Exception(error, "FT_Bitmap_Copy");
	 library = other.library;
      }
      Bitmap& operator=(Bitmap&& other) noexcept {
	 swap(other);
	 return *this;
      }
      void swap(Bitmap& other) noexcept {
	 std::swap(library, other.library);
	 std::swap(bitmap, other.bitmap);
      }
      unsigned int get_num_rows() const {
	 return bitmap.rows;
      }
      unsigned int get_width() const {
	 return bitmap.width;
      }
      unsigned int get_pitch() const {
	 return bitmap.pitch;
      }
      unsigned int bits_per_pixel() const {
	 switch (bitmap.pixel_mode) {
	    case FT_PIXEL_MODE_MONO: return 1;
	    case FT_PIXEL_MODE_GRAY: return 8;
	    case FT_PIXEL_MODE_GRAY2: return 2;
	    case FT_PIXEL_MODE_GRAY4: return 4;
	    case FT_PIXEL_MODE_LCD: return 8 * 3;
	    case FT_PIXEL_MODE_LCD_V: return 8 * 3;
	    case FT_PIXEL_MODE_BGRA: return 8 * 4;
	    default: return 0;
	 }
      }
      unsigned int operator()(unsigned int row, unsigned int col) const {
	 assert(row < bitmap.rows && col < bitmap.width);
	 unsigned int bits = bits_per_pixel();
	 assert(bits <= 8 && 8 % bits == 0);
	 unsigned int offset = bits * col;
	 unsigned int val = bitmap.buffer[row * bitmap.pitch + offset / 8];
	 val >>= 7 - offset % 8;
	 val &= (1<<bits)-1;
	 return val;
      }
      const unsigned char* addr(unsigned int row, unsigned int col) const {
	 assert(row < bitmap.rows && col < bitmap.width);
	 unsigned int bits = bits_per_pixel();
	 assert(bits >= 8 && bits % 8 == 0);
	 return bitmap.buffer + row * bitmap.pitch + bits / 8 * col;
      }
   private:
      FT_Library library;
      FT_Bitmap bitmap;
};

} // namespace freetype

#endif
