#ifndef FTPP_FACE_H
#define FTPP_FACE_H

#include <cassert>
#include <utility>
#include <freetype2/ft2build.h>
#include FT_FREETYPE_H
#include <ftpp/exception.hpp>
#include <ftpp/glyph.hpp>
#include <ftpp/bitmap.hpp>
#include <ftpp/f26dot6.hpp>
#include <ftpp/vector.hpp>
#include <ftpp/bbox.hpp>

namespace freetype {

class Face {
   public:
      Face() noexcept : face(nullptr), face_index(0), glyph_loaded(false) {
      }
      Face(FT_Library library, const char* filename,
	    FT_Long face_index = 0) :
	       library(library), face_index(face_index), glyph_loaded(false) {
	 FT_Error error = FT_New_Face(library, filename, face_index, &face);
	 if (error) throw Exception(error, "FT_New_Face", filename);
      }
      Face(Face&& other) noexcept {
	 swap(other);
      }
      ~Face() {
	 if (face) FT_Done_Face(face);
      }
      Face& operator=(const Face& other) = delete;
      Face& operator=(Face&& other) noexcept {
	 swap(other);
	 return *this;
      }
      void swap(Face& other) noexcept {
	 std::swap(face, other.face);
	 std::swap(face_index, other.face_index);
	 std::swap(glyph_loaded, other.glyph_loaded);
      }

      /* properties of a typeface */
      FT_Long get_num_faces() const {
	 assert(face_index == 0);
	 return face->num_faces;
      }
      FT_Long get_num_glyphs() const {
	 return face->num_glyphs;
      }
      FT_Long get_face_flags() const {
	 return face->face_flags;
      }
      const char* get_family_name() const {
	 return face->family_name;
      }
      const char* get_style_name() const {
	 return face->style_name;
      }
      bool scalable() const {
	 return FT_IS_SCALABLE(face);
      }
      bool fixed_width() const {
	 return FT_IS_FIXED_WIDTH(face);
      }
      bool has_kerning() const {
	 return FT_HAS_KERNING(face);
      }
      FT_UShort get_units_per_EM() const {
	 return face->units_per_EM;
      }
      BoundingBox get_bbox() const {
	 return BoundingBox(face->bbox);
      }
      FT_Short get_ascender() const {
	 return face->ascender;
      }
      FT_Short get_descender() const {
	 return face->descender;
      }
      FT_Short get_height() const {
	 return face->height;
      }

      void set_char_size(f26dot6 width, f26dot6 height,
	    FT_UInt horizontal_resolution,
	    FT_UInt vertical_resolution) {
	 FT_Error error = FT_Set_Char_Size(face,
	    width.get(), height.get(),
	    horizontal_resolution, vertical_resolution);
	 if (error) throw Exception(error, "FT_Set_Char_Size");
      }

      FT_UInt get_glyph_index(FT_ULong charcode) const {
	 return FT_Get_Char_Index(face, charcode);
      }

      Vector get_kerning(FT_UInt left_glyph, FT_UInt right_glyph,
	    FT_UInt kern_mode = FT_KERNING_DEFAULT) {
	 FT_Vector kerning_vector;
	 FT_Error error = FT_Get_Kerning(face, left_glyph, right_glyph,
	    kern_mode, &kerning_vector);
	 if (error) throw Exception(error, "FT_Get_Kerning");
	 return kerning_vector;
      }

      /* glyph operators working on the glyph slot within a face */
      void load_glyph(FT_UInt glyph_index,
	    FT_Int32 load_flags = FT_LOAD_DEFAULT) {
	 FT_Error error = FT_Load_Glyph(face, glyph_index, load_flags);
	 if (error) throw Exception(error, "FT_Load_Glyph");
	 glyph_loaded = true;
      }
      void load_char(FT_ULong charcode,
	    FT_Int32 load_flags = FT_LOAD_DEFAULT) {
	 FT_Error error = FT_Load_Char(face, charcode, load_flags);
	 if (error) throw Exception(error, "FT_Load_Glyph");
	 glyph_loaded = true;
      }
      Glyph get_glyph() const {
	 assert(glyph_loaded);
	 return Glyph(library, face->glyph);
      }
      Glyph get_glyph(FT_UInt glyph_index,
	    FT_Int32 load_flags = FT_LOAD_DEFAULT) {
	 load_glyph(glyph_index, load_flags);
	 return get_glyph();
      }
      void render_glyph(FT_Render_Mode render_mode = FT_RENDER_MODE_NORMAL) {
	 assert(glyph_loaded);
	 FT_Error error = FT_Render_Glyph(face->glyph, render_mode);
	 if (error) throw Exception(error, "FT_Render_Glyph");
      }
      const Bitmap get_bitmap() const {
	 return Bitmap(library, &face->glyph->bitmap);
      }
      Vector get_advance() const {
	 return Vector(face->glyph->advance);
      }
      FT_Int get_left() const {
	 return face->glyph->bitmap_left;
      }
      FT_Int get_top() const {
	 return face->glyph->bitmap_top;
      }
   private:
      FT_Library library;
      FT_Face face;
      FT_Long face_index;
      bool glyph_loaded;
};

} // namespace freetype

#endif
